"""
This is the rlrom module.

It contains methods for robust online monitoring of reinforcement learning models.
"""

from .envs import supported_envs, supported_models, cfg_envs

__all__ = ['supported_envs', 'supported_models', 'cfg_envs']



from bokeh.models.annotations import Title
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show
from bokeh.palettes import Dark2_5 as palette
import itertools
import re
import matplotlib.pyplot as plt
import rlrom.utils as utils


def get_layout_from_string(signals_layout):
    out = []
    # split signals string wrt linebreaks first
    signals_rows = signals_layout.splitlines()

    # then strip and split wrt commas
    for line in signals_rows:        
        if line.strip() == '':
            continue
        else:
            out_row = re.findall(r'\b[A-Za-z_][A-Za-z0-9_]*\b(?:\([^)]*\))?', line)            
            out.append(out_row)        

    return out            


def get_fig(envs, signals_layout, tr_idx=0):
# plots stuff in bokeh from a list of environments
         
    if not isinstance(envs, list):
         envs= [envs]
    current_env= envs[tr_idx]                                                 
    lay = get_layout_from_string(signals_layout)
    status = "Plot ok. Hit reset on top right if not visible."            

    #f= figure(height=200)
    figs = []
    colors = itertools.cycle(palette)    
    
    for signal_list in enumerate(lay):
        f=None
        for signal in signal_list[1]:                
            #try: 
                color=colors.__next__()                                        
                #print(signal.strip())
                if signal.strip().startswith("set_trace_idx(") or signal.strip().startswith("_tr("):            
                    tr_idx = int(signal.split('(')[1][:-1])
                    current_env= envs[tr_idx]                                                 
                else: 
                    if f is None:
                        if figs == []:
                            f = figure(height=200)
                        else:
                            f = figure(height=200, x_range=figs[0][0].x_range)
                        figs.append([f])
                    
                    ttime = current_env.get_time()                        
                    labl = signal                                                
                    if len(envs)>1:                                                            
                        labl += ', trace_idx='+str(tr_idx)

                    sig_values, sig_type = current_env.get_values_from_str(signal)
                    if sig_type == 'val':
                        f.scatter(ttime, sig_values, legend_label=labl, color=color)
                        f.line(ttime, sig_values, legend_label=labl, color=color)
                    elif sig_type == 'rob':
                        f.step(ttime, sig_values, legend_label=labl, color=color)                        
                        
                    elif sig_type == 'sat':
                        f.step(ttime, sig_values, legend_label=labl, color=color)                        
                                            
            #except:
            #     status = "Warning: error getting values for " + signal
    fig = gridplot(figs, sizing_mode='stretch_width')        
        
    return fig, status

def plot_training_logs(data, ax=None, label=None, color='gray', linestyle='-'):
# assumes data is a list of sync logs with steps and values fields    
    
    steps = data[0].get('steps')
    mean_values = utils.get_mean_values(data)
    min_values = utils.get_lower_values(data)
    max_values = utils.get_upper_values(data)    

    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 4))
        ax.grid(True)
    if label is None:    
        ax.plot(steps, mean_values, color=color,linestyle=linestyle)
    else:
        ax.plot(steps, mean_values,label=label, color=color,linestyle=linestyle)

    ax.plot(steps, min_values, color=color, linestyle=linestyle)
    ax.plot(steps, max_values, color=color,linestyle=linestyle)

    ax.fill_between(steps, min_values, max_values, color=color, alpha=0.25, label='Min-Max Range')
    ax.legend()
    ax.set_xlabel('Training Steps')
    ax.set_ylabel('Average cumulated reward')


    return ax


from stable_baselines3 import PPO #,A2C,SAC,TD3,DQN,DDPG
import rlrom.utils as utils
from rlrom.testers import RLTester
from stable_baselines3.common.env_util import make_vec_env
from stable_baselines3.common.vec_env import SubprocVecEnv
from stable_baselines3.common.callbacks import BaseCallback, CallbackList
import numpy as np
import yaml
import time


class STLWrapperCallback(BaseCallback):
  def __init__(self, verbose=0, cfg_main=dict()):
    super().__init__(verbose)
    self.tester = RLTester(cfg_main, render_mode=None)
    self.tester.init_env()    
  

  def _on_step(self):
    return True
    
  def _on_rollout_end(self):
    # Access the rollout buffer
    buffer = self.model.rollout_buffer
    episodes = utils.get_episodes_from_rollout(buffer)
    num_ep = len(episodes)
    
                
    res_rew = dict()
    res_rew_f_list = []    
    res_eval_f_list = []
    
    ep_lens = []
    for i in range(0,num_ep):        
      #print('----------------------------------------------------')
      #print('EPISODE', i, end=' ')            
      self.tester.env.set_episode_data(episodes[i])      
      # eval reward formulas and eval formulas
      res_rew, res_all_ep, res_rew_f_list, res_eval_f_list = self.tester.env.eval_episode(episodes[i],res=res_rew,res_rew_f_list=res_rew_f_list, res_eval_f_list=res_eval_f_list)
      rewards = episodes[i]['rewards']
      ep_lens.append(len(rewards))
      #print()
    my_mean_ep_len =   np.array(ep_lens).mean()
    print('Number of episodes:', num_ep, 'mean ep len:', my_mean_ep_len )
    self.logger.record('rollout/my_mean_ep_len', my_mean_ep_len)
      

    for metric_name, metric_value in res_all_ep['basics'].items():
      log_name = 'basics/'+metric_name    
      self.logger.record(log_name, metric_value)
    for f_name, f_value in res_all_ep['eval_formulas'].items():    
      for metric_name, metric_value in f_value.items():
        log_name = 'eval_f/'+f_name+'/'+metric_name 
        self.logger.record(log_name, metric_value)
    for f_name, f_value in res_all_ep['reward_formulas'].items():    
      for metric_name, metric_value in f_value.items():
        log_name = 'rew_f/'+f_name+'/'+metric_name 
        self.logger.record(log_name, metric_value)

    return True

class RLTrainer:
  def __init__(self, cfg):    
    self.cfg = utils.load_cfg(cfg)    
    self.cfg_train = self.cfg.get('cfg_train', {})
    pass   

  def train(self,make_env):
    
    cfg_algo = self.cfg_train.get('algo')
    model_name = self.cfg_train.get('model_name')
    
    if cfg_algo is not None:       
       if cfg_algo.get('ppo') is not None:                     
          cfg_ppo = cfg_algo.get('ppo')
          print('Training  with PPO...',cfg_ppo )
          callbacks = CallbackList([        
          STLWrapperCallback(verbose=1, cfg_main=self.cfg)        
          ])
          model = self.train_ppo(cfg_ppo, make_env, model_name,callbacks)
    # Saving the agent
    model_name, cfg_name = utils.get_model_fullpath(self.cfg)
    model.save(model_name) #TODO try except 
    with open(cfg_name,'w') as f:
         yaml.safe_dump(self.cfg, f)


    return model

  def train_ppo(self,cfg, make_env,model_name, callbacks):
      
    # hyperparams, training configuration      
    n_envs = cfg.get('n_envs',1)
    batch_size = cfg.get('batch_size',128)
    neurons = cfg.get('neurons',128)
    learning_rate = float(cfg.get('learning_rate', '5e-4'))
    total_timesteps = cfg.get('total_timesteps',1000)
    cfg_tb = cfg.get('tensorboard',dict()) 
    tb_dir, tb_prefix= self.get_tb_dir(cfg_tb,model_name)
    
    policy_kwargs = dict(
      #activation_fn=th.nn.ReLU,
      net_arch=dict(pi=[neurons, neurons], qf=[neurons, neurons])
    )
    if n_envs>1:
       env = make_vec_env(make_env, n_envs=n_envs, vec_env_cls=SubprocVecEnv)    
    else:
       env = make_env()

    # Instantiate model
    model = PPO("MlpPolicy",env,
    device='cpu',
    policy_kwargs=policy_kwargs,
    n_steps=batch_size * 12 // n_envs,
    batch_size=batch_size,
    n_epochs=10,
    learning_rate=learning_rate,
    gamma=0.9,
    verbose=1,
    tensorboard_log=tb_dir
    )
 
    # Train the agent
    model.learn(
      total_timesteps=total_timesteps,
      callback = callbacks,
      tb_log_name=tb_prefix,
      progress_bar=True
    )
    
    return model

  def get_tb_dir(self, cfg, model_name):
    tb_dir = cfg.get('tb_path','./tb_logs')
    tb_prefix =  f"{model_name}_{int(time.time())}"
    return tb_dir, tb_prefix


"""Wrapper for transforming the reward."""
import gymnasium as gym
from gymnasium import spaces
#from tensorflow.python.ops.numpy_ops.np_dtypes import float32
from collections import Counter
import numpy as np
import math

class RewardMachine(gym.Wrapper):
    """Transform the reward via reward machine.

    Warning:
        If the base environment specifies a reward range which is not invariant under :attr:`f`, the :attr:`reward_range` of the wrapped environment will be incorrect.

    Example:
        >>> import gymnasium as gym
        >>> from gymnasium.wrappers import TransformReward
        >>> env = gym.make("CartPole-v1")
        >>> env = RewardMachine(env)
        >>> _ = env.reset()
        >>> observation, reward, terminated, truncated, info = env.step(env.action_space.sample())
    """

    def __init__(self, env, rm_filename):
        """Initialize the :class:`RewardMachine` wrapper with an environment and the file that contains the reward machine.

        Args:
            env: The environment to apply the wrapper
            rm_filename: The location and the filename of the reward machine
        """
        gym.Wrapper.__init__(self, env)
        self.env = env
        self.timestep = 0
        self.rm_list, self.num_states , self.ut = self._load_reward_machine(f"./{rm_filename}")
        self.u_in = 0
        self.task = 0

    def step(self, action):
        """Modify the step function
        :param action: same as the original step
        :return: observation with the reward machine state and robustness, and the new reward value from the
        reward machine, terminated, truncated, info.
        """
        obs, reward, terminated, truncated, info = self.env.step(action)
        self.timestep += 1
        reward_env = reward
        u_in = self.u_in
        self.u_in, rm_reward = self.get_rm_transition(u_in, action, self.timestep, obs[0],reward_env)
        
        if terminated: self.env.reset()
        return obs, rm_reward, terminated, truncated, info

    def reset(self, **kwargs):
        self.timestep = 0
        self.u_in = 0
        return self.env.reset(**kwargs)

    def render(self):
        return self.env.render()

    def close(self):
        return self.env.close()

    def seed(self, seed):
        return self.env.reset(seed=seed)

    def get_rm_transition(self, u_in, action, ts, x, reward_env):
        # Evaluate the new state given the input state with the true specification in the reward machine
        u_out = 0 ; reward = reward_env
        for transition in self.rm_list:
            if transition[3] == "reward_env":
                transition[3] = reward_env
            if u_in == transition[0]:
                p = transition[2]
                if eval(p): # boolean variable
                    u_out = transition[1]
                    reward = float(transition[3])
                    break
                else:
                    u_out = u_in
        return u_out, reward

    def _load_reward_machine(self, file):
        f_copy = open(file)
        len_file = len(f_copy.readlines())
        f = open(file)
        self.u0 = int(f.readline()[0]) # Initial state. First line of txt file
        terminal_state = f.readline().split('[')[1].split(']')[0] # Terminal state(s). Second line of txt file
        terminal_state = terminal_state.split(",")
        try:
            self.ut = [int(i) for i in terminal_state]
        except ValueError:
            self.ut = None # If there is no terminal state

        rm_list = []
        for line in range(len_file - 2): # Rest of the lines in txt file --> [ui, uo, dnf, reward_env]
            rm_line = f.readline().replace("\n", "").split(",")
            rm_list.append(rm_line)
        num_states = [] # Count number of different states in the RM and change the type str of the RM states to int
        for U in rm_list:
            U[:2] = [int(u) for u in U[:2]]
            num_states += U[:2]
        num_states = len(Counter(num_states).keys())
        #returns a list of all possible transitions and the number of states of the reward machine
        return rm_list, num_states, self.ut


MAX_NUM_STATE = 100 # Max number of state for reward machine - we might want do reconsider if we go into composing and stuff
class RewardMachineObservation(gym.ObservationWrapper):
    def __init__(self, env):
        gym.ObservationWrapper.__init__(self, env)
        self.u0 = 0
        self.ut = None
        
        low = self.observation_space.__getattribute__("low") 
        high = self.observation_space.__getattribute__("high")
        
        self.observation_space = spaces.Box(np.append(low, self.u0), 
                                            np.append(high, MAX_NUM_STATE),        
                                            dtype=np.float32)
        
        #self.observation_space = spaces.Box(np.append(low, np.array([self.u0, -1000])), 
        #                                    np.append(high,np.array([MAX_NUM_STATE, 1000])),        
        #                                    dtype=np.float32)
        

    def observation(self, observation):
        # observation with rm state 
        observation = np.append(observation , self.get_wrapper_attr('u_in'))

        # Observation only rm state and dummy 0. after
        # observation = np.append(observation , np.array([self.get_wrapper_attr('u_in'),0.]))
        
        return observation
    




from stable_baselines3 import PPO,A2C,SAC,TD3,DQN,DDPG
from sb3_contrib import TRPO, QRDQN
from huggingface_hub import HfApi
from huggingface_sb3 import load_from_hub
from huggingface_sb3.naming_schemes import EnvironmentName, ModelName, ModelRepoId
import re
import yaml
import os
import numpy as np
from tensorboard.backend.event_processing import event_accumulator

# helper function to concat new values in a dict field array
def append_to_field_array(res, metric, val):
    vals = res.get(metric,None)
    if vals is None:
        res[metric]=np.array([val])
    else:
        vals = np.atleast_1d(vals)
        vals = np.append(vals,val)
        res[metric]= vals
    return res

def list_folders(folder, filter=''):
    try:
        # List all items in the given directory
        items = os.listdir(folder)
        # Filter out only the directories
        folders = [os.path.join(folder, item) for item in items 
                   if (os.path.isdir(os.path.join(folder, item)) and
                       filter in item)]
        
        return folders
    except Exception as e:
        print(f"An error occurred: {e}")
        return []
    
def tb_extract_from_tag(file_path_list, tag='rollout/ep_rew_mean'):
# return a list of data dict with fields steps and values        

    if not(isinstance(file_path_list,list)):
        file_path_list = [file_path_list]
    
    all_data = []
    for file_path in file_path_list:
        l = os.listdir(file_path)
        event_file = file_path+'/'+l[0]    
    
        # Initialize the event accumulator
        ea = event_accumulator.EventAccumulator(event_file)
        ea.Reload()

        # Extract scalar values for the specified tag        
        data = dict()
        if tag in ea.Tags()['scalars']:
            scalar_events = ea.Scalars(tag)
            data["steps"] = [event.step for event in scalar_events]
            data["values"] = [event.value for event in scalar_events]            
        
        all_data.append(data)

    return all_data    

def get_upper_values(all_data):
    # Assumes all_data in sync (i.e. same steps)

    all_values = []
    for v in all_data:
        all_values.append(v.get('values'))

    return np.max(all_values,axis=0)        

def get_lower_values(all_data):
# Assumes all_data in sync (i.e. same steps)

    all_values = []
    for v in all_data:
        all_values.append(v.get('values'))

    return np.min(all_values, axis=0)        

def get_mean_values(all_data):
    # Assumes all_data in sync (i.e. same steps)

    all_values = []
    for v in all_data:
        all_values.append(v.get('values'))

    return np.mean(all_values,axis=0)        

# load cfg recursively 
def load_cfg(cfg, verbose=1):
    def recursive_load(cfg):
        for key, value in cfg.items():
            #print('reading', key, 'with value', value)
            if isinstance(value, str) and value.endswith('.yml'):
                if verbose>=1:
                    print('loading field [', key, '] from YAML file [', value, ']')
                    with open(value, 'r') as f:                        
                        cfg[key] = recursive_load(yaml.safe_load(f))                
                else:
                    cfg[key] = value
                    print('WARNING: file', value,'not found!')
            elif isinstance(value, str) and value.endswith('.stl'):
                if verbose>=1:
                    print('loading field [', key, '] from STL file [', value, ']')            
                with open(value,'r') as F: 
                    cfg[key]= F.read()
            elif isinstance(value, dict):
                cfg[key]= recursive_load(value)

        return cfg

    if isinstance(cfg, str) and os.path.exists(cfg):
        with open(cfg, 'r') as f:
            cfg= yaml.safe_load(f)
    elif not isinstance(cfg, dict): 
        raise TypeError(f"Expected file name or dict.")
    
    return recursive_load(cfg)

def get_model_fullpath(cfg):
    # returns absolute path for model, as well as for yaml config (may not exist yet)
    # The yml file (second output), if it exists, contains the full configuration used 
    # to train the model
    model_path = cfg.get('model_path', './models')
    model_name = cfg.get('model_name', 'random')
    full_path = os.path.join(model_path, model_name+'.zip')
    
    if model_name=='random':
        if os.path.exists(full_path):
            print(f"WARNING: Somehow a model was named 'random' (path: {full_path}). Rename it if you actually want to use it.")        
        full_path = 'random'
        cfg_full_path = None
    else:
        if not os.path.exists(full_path):
            print(f"WARNING: Path does not exist: {full_path}")
        else:
            full_path= os.path.abspath(full_path)
            cfg_full_path = full_path.replace('.zip', '.yml')
    
    return full_path, cfg_full_path

def find_model(env_name):
    return find_huggingface_models(env_name)

def find_huggingface_models(env_name, repo_contains='', algo=''):
    api = HfApi()
    models_iter = api.list_models(tags=env_name)
    
    models = []
    models_ids = []
    for model in models_iter:
        model_id = model.modelId
        if repo_contains.lower() in model_id.lower() and algo.lower() in model_id.lower(): 
            models.append(model)
            models_ids.append(model_id)
   
    return models, [model.modelId for model in models]

def load_model(env_name, repo_id=None, filename=None):

    if repo_id is None and filename is None:
        filename=env_name

    model = None
    # checks if filename point to a valid file
    if filename is not None:
        try:
            with open(filename, 'r') as f:
                pass
            
            # try loading with PPO, A2C, SAC, TD3, DQN, QRDQN, DDPG, TRPO
            try:
                model= PPO.load(filename)
                print("loading PPO model succeeded")                
                return model
            except:
                print("loading PPO model failed")
                pass

            try:
                model= A2C.load(filename)
                print("loading A2C model succeeded")                
                return model
            except:
                print("loading A2C model failed")
                pass

            try:
                model= SAC.load(filename)
                print("loading SAC model succeeded")                
                return model
            except:
                print("loading SAC model failed")                
                pass

            try:
                model= TD3.load(filename)
                print("loading TD3 model succeeded")                                                
                return model
            except:
                print("loading TD3 model failed")
                pass

            try:                
                model= DQN.load(filename)
                print("loading DQN model succeeded")                
                return model
            except:
                print("loading DQN model failed")
                pass    

            try:
                model= QRDQN.load(filename)
                print("loading QRDQN model succeeded")                
                return model
            except:
                print("loading QRDQN model failed")
                pass

            try:
                model= DDPG.load(filename)
                print("loading DDPG model succeeded")                                    
                return model
            except:
                print("loading DDPG model failed")
                pass
            try:
                model= TRPO.load(filename)
                print("loading TRPO model succeeded")                
                return model
            except:
                print("loading TRPO model failed")
                pass            
        except FileNotFoundError:
            print("File not found",filename)            

    if repo_id is not None:
        if 'ppo' in repo_id:
            model = load_ppo_model(env_name, repo_id, filename=filename)
        elif 'a2c' in repo_id:
            model = load_a2c_model(env_name, repo_id, filename=filename)
        elif 'sac' in repo_id:
            model = load_sac_model(env_name, repo_id, filename=filename)
        elif 'td3' in repo_id:
            model = load_td3_model(env_name, repo_id, filename=filename)
        elif 'dqn' in repo_id:
            model = load_dqn_model(env_name, repo_id, filename=filename)
        elif 'qrdqn' in repo_id:
            model = load_qrdqn_model(env_name, repo_id, filename=filename)
        elif 'ddpg' in repo_id:
            model = load_ddpg_model(env_name, repo_id, filename=filename)
        elif 'trpo' in repo_id:
            model = load_trpo_model(env_name, repo_id, filename=filename)
        else:
            model = None
    return model

def get_episodes_from_rollout(buffer):
# Takes a rollout buffer as produced by PPO and returns a list of complete episodes     
    episodes = []
    for env_idx in range(buffer.n_envs):
        env_dones = buffer.episode_starts[:, env_idx] # dones flags for each steps

        sz = np.shape(buffer.observations)        
        if sz[0]==buffer.buffer_size:
            env_obs = buffer.observations[:, env_idx]  # All steps for this env
        else:
            start_idx_env = env_idx*buffer.buffer_size 
            end_idx_env = env_idx*buffer.buffer_size + buffer.buffer_size
            env_obs = buffer.observations[start_idx_env:end_idx_env]  # All steps for this env
        
        sz = np.shape(buffer.actions)        
        if sz[0]==buffer.buffer_size:            
            env_actions = buffer.actions[:, env_idx]  # All actions for this env            
        else:
            start_idx_env = env_idx*buffer.buffer_size 
            end_idx_env = env_idx*buffer.buffer_size + buffer.buffer_size
            env_actions = buffer.actions[start_idx_env:end_idx_env]  # All actions for this env

        sz = np.shape(buffer.rewards)        
        if sz[0]==buffer.buffer_size:            
            env_rewards = buffer.rewards[:, env_idx]  # All rewards for this env            
        else:
            start_idx_env = env_idx*buffer.buffer_size 
            end_idx_env = env_idx*buffer.buffer_size + buffer.buffer_size
            env_rewards = buffer.rewards[start_idx_env:end_idx_env]  # All actions for this env
            
        
        # Split into episodes based on done flags
        episode_start = 0                
        for step in range(len(env_dones)):
            episode = dict()
            if env_dones[step] and step > 0:  # found episode boundary
                episode['observations'] = env_obs[episode_start:step]                
                episode['actions'] = env_actions[episode_start:step]                
                episode['rewards'] = env_rewards[episode_start:step]                
                episode['dones'] = env_dones[episode_start:step]                
                episodes.append(episode)
                episode_start = step                        
        # note we only want complete episodes, so we drop the last observations for each batch, if they don't end with a done

    return episodes

def parse_signal_spec(signal):
    # extract sig_name and args from signal_name(args)
    signal = signal.split('(')
    sig_name = signal[0]
    if len(signal) == 1:
        args = []
    else:
        args = [arg.strip() for arg in signal[1][:-1].split(',')]
    return sig_name, args

def get_formulas(specs):
    # regular expression matching id variable in the specs at the beginning of a line followed by :=
    # then the rest of the line
    regex = r"^\s*\b([a-zA-Z_][a-zA-Z0-9_]*)\b\s*:="

    # find all variable id in the stl_string
    formulas = re.findall(regex, specs, re.MULTILINE)
    return formulas

def parse_integer_set_spec(str):
    sp_str = str.split(',')
    idx_out = []
    for s in sp_str:
        s = s.strip()
        if s.isdigit():
            idx_out.append(int(s))
        else:
            [l, h] = s.split(':')
            if l.isdigit() and h.isdigit():
                range_idx = [ idx for idx in range(int(l),int(h)+1) ]
                idx_out = idx_out+range_idx
    return idx_out                

def get_symmetric_max(sig):
    npsig= np.array(sig)
    max_pos = npsig.max()
    min_neg = -npsig.min()
    return max(max_pos,min_neg)

# Auxiliary load functions        
def load_ppo_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('ppo', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)
    custom_objects = {
            "learning_rate": 0.0,
            "lr_schedule": lambda _: 0.0,
            "clip_range": lambda _: 0.0,
    }
    model = PPO.load(checkpoint, custom_objects=custom_objects, print_system_info=True)
    return model

def load_a2c_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('a2c', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)  
    model = A2C.load(checkpoint, print_system_info=True)
    return model

def load_sac_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('sac', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)  
    model = SAC.load(checkpoint, print_system_info=True)
    return model

def load_td3_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('td3', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)  
    model = TD3.load(checkpoint, print_system_info=True)
    return model

def load_dqn_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('dqn', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)  
    model = DQN.load(checkpoint, print_system_info=True)
    return model

def load_qrdqn_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('dqn', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)  
    model = QRDQN.load(checkpoint, print_system_info=True)
    return model

def load_ddpg_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('ddpg', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)
    model = DDPG.load(checkpoint, print_system_info=True)
    return model

def load_trpo_model(env_name, repo_id, filename=None):
    if filename is None:
        filename = ModelName('trpo', env_name)+'.zip'
    checkpoint = load_from_hub(repo_id=repo_id, filename=filename)
    model = TRPO.load(checkpoint, print_system_info=True)
    return model



import gymnasium as gym
from gymnasium import spaces
import numpy as np
import stlrom
import matplotlib.pyplot as plt
from bokeh.models.annotations import Title
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show
from bokeh.palettes import Dark2_5 as palette
from rlrom.utils import append_to_field_array as add_metric

class STLWrapper(gym.Wrapper):
    
    def __init__(self,env,
                 stl_driver, 
                 signals_map={},
                 obs_formulas={},  
                 reward_formulas={},
                 eval_formulas={},
                 end_formulas={},
                 BigM=None
                 ):
        gym.Wrapper.__init__(self, env)
        self.env = env
        self.real_time_step = 1
        self.time_step = 0     # integer current time step 
        self.current_time = 0  # real time (=time_step*real_time_step)  for formula evaluation
        self.stl_driver = stl_driver                        
        self.obs_formulas = obs_formulas
        self.reward_formulas = reward_formulas
        self.eval_formulas = eval_formulas
        self.end_formulas= end_formulas
        self.episode={'observations':[], 'actions':[],'rewards':[], 'dones':[]}

        self.signals_map={}
        if signals_map=={}:
            # assumes 1 action and n obs
            signals = stl_driver.get_signals_names().split()
            i_sig=0

            for sig in signals:
                if i_sig==0:
                    self.signals_map[sig] = 'action'
                elif i_sig<len(signals)-1:    
                    self.signals_map[sig] = f'obs[{i_sig-1}]'
                else:
                    self.signals_map[sig] = 'reward'
                i_sig+=1

        elif type(signals_map)==dict: 
            if 'action_names' in signals_map:
                for a_name,a_ref in signals_map['action_names'].items():                    
                    self.signals_map[a_name]=a_ref            
            if 'obs_names' in signals_map:
                for o_name,o_ref in signals_map['obs_names'].items():                     
                    self.signals_map[o_name]=o_ref
            if 'aux_sig_names' in signals_map:
                for o_name,o_ref in signals_map['aux_sig_names'].items():                     
                    self.signals_map[o_name]=o_ref

            self.signals_map['reward'] = 'reward'
        else: # assumes all is fine (TODO? catch bad signals_map here)    
            self.signals_map=signals_map
        
        low = self.observation_space.__getattribute__("low") 
        high = self.observation_space.__getattribute__("high")
        if BigM is None:
            BigM = stlrom.Signal.get_BigM()
        self.observation_space = spaces.Box(np.append(low,  [-BigM]*len(self.obs_formulas)), 
                                            np.append(high, [BigM]*len(self.obs_formulas)),        
                                            dtype=np.float32)
        
        idx_obs_f = self.observation_space.shape[0]-len(self.obs_formulas) # adding mapping from stl signal to obs array, now flat 
        for f_name, f_opt in obs_formulas.items():             
            f_hor = f_opt.get('past_horizon',0)
            obs_name = 'obs_'+f_name+'_hor_'+str(f_hor)
            obs_name = f_opt.get('obs_name', obs_name)
            ref_in_obs = 'obs['+str(idx_obs_f)+']'
            self.signals_map[obs_name]= ref_in_obs
            idx_obs_f +=1

    def step(self, action):
    
        # steps the wrapped env
        obs, reward, terminated, truncated, info = self.env.step(action)                
        
        # store wrapped obs
        self.wrapped_obs = obs
                
        # collect the sample for monitoring 
        s = self.get_sample(self.prev_obs, action, reward) 
        
        # add sample and compute robustness
        self.stl_driver.add_sample(s)        
        
        idx_formula = 0
        rob = [0]*len(self.obs_formulas)
        for f_name, f_opt in self.obs_formulas.items():             
            robs_f,_ = self.eval_formula_cfg(f_name,f_opt)        
            rob[idx_formula] = robs_f # forget about low and high rob for now
            idx_formula+=1     
         
        for f_name, f_opt in self.end_formulas.items():             
            _, res = self.eval_formula_cfg(f_name,f_opt)          
            if res['lower_rob'] > 0:
                print('Episode terminated because of formula', f_name)
                terminated = True

        new_reward = reward                         
        # add stl robustness to reward
        for f_name, f_opt in self.reward_formulas.items():             
            robs_f,_ = self.eval_formula_cfg(f_name,f_opt)
            w = f_opt.get('weight',1)        
            new_reward += w*robs_f   
        
        # update current time
        self.time_step += 1
        self.current_time += self.real_time_step
        
        # return obs with added robustness
        new_obs = np.append(obs, rob)
                        
        self.prev_obs = new_obs

        self.episode['observations'].append(new_obs)               
        self.episode['actions'].append(action)
        self.episode['rewards_wrapped'].append(reward)
        self.episode['rewards'].append(new_reward)
        self.episode['dones'].append(terminated)
            

        if terminated: 
            self.env.reset()

        return new_obs, new_reward, terminated, truncated, info

    def get_sample(self,obs,action,reward):        
        # get a sample for stl driver, converts obs, action,reward into (t, signals_values)
        s = np.zeros(len(self.signals_map)+1-len(self.obs_formulas))
        s[0] = self.current_time
        i_sig = 0
        for key, value in self.signals_map.items():
            i_sig = i_sig+1
            if i_sig>len(s)-1:
                break
            #print(key, value)
            s[i_sig] = eval(value)

        return s

    def reset(self, **kwargs):        
        self.time_step = 0
        self.current_time = 0
        obs0, info = self.env.reset(**kwargs)
        self.wrapped_obs = obs0        
        robs0 = self.reset_monitor()
        obs = np.append(obs0, robs0)
        self.prev_obs = obs
        self.episode={'observations':[], 'actions':[],'rewards':[], 'rewards_wrapped':[],'dones':[]}        
        return obs, info

    def reset_monitor(self):        
        self.stl_driver.data = [] 
        return [0]*len(self.obs_formulas) 
    
    def render(self):
        return self.env.render()

    def close(self):
        return self.env.close()

    def seed(self, seed):
        return self.env.reset(seed=seed)
    
    def plot_signal(self, signal, fig=None,label=None,  color=None, online=False, past_horizon=0, linestyle='-', booleanize=False):
    # signal should be part of the "signal" declaration or a valid formula id 
     
        if self.stl_driver.data == []:
            raise ValueError("No data to plot.")
                 
        time = self.get_time()
        sig_values = self.get_sig(signal)
        if sig_values is None:
            if signal in self.formulas:
                sig_values = self.get_rob(signal, online=online,past_horizon=past_horizon)
                signal_index = self.formulas.index(signal)+len(self.signals_map)        
            elif isinstance(signal, np.ndarray) and signal.shape == (len(self.get_time()),):
                sig_values = signal
            elif isinstance(signal, stlrom.Signal):
                pass
            else:
                try:
                    sig_values = self.get_rob(signal, online=online,past_horizon=past_horizon)
                except Exception as e:
                    raise ValueError(f"Name '{signal}' not in signals_map nor in parsed formulas")

        if booleanize:
            sig_values = (sig_values >0).astype(int)

        if fig is None:
             fig = figure(height=200)

        fig.set_xlabel('Time')
        fig.grid(True)

        fig.step(time, sig_values)
        if color is None:
            l, = fig.step(time, sig_values, label=label,linestyle=linestyle)
            color = l.get_color()
        else:
            l = fig.step(time, sig_values, color=color,linestyle=linestyle)
        
        if label is None:
            label=signal
            
        l.set_label(label)
        fig.legend()

        return fig

    def get_time(self):
        if self.stl_driver.data == []:
            raise ValueError("No data to plot.")
        
        return [s[0] for s in self.stl_driver.data]
        
    def get_sig(self, sig_name):
        sig = None        
        sig_expr = self.signals_map.get(sig_name,[])
        if sig_expr != []:                     
            observations = self.episode['observations']
            actions = self.episode['actions']
            rewards = self.episode['rewards']            
            rewards_wrapped = self.episode.get('rewards_wrapped',rewards) 
            step = 0
            sig=[]            
            while step<len(observations):
                obs = observations[step]
                action = actions[step]
                reward = rewards[step]
                reward_wrapped = rewards_wrapped[step]                
                sig.append(eval(sig_expr))                
                step +=1
        return sig

    def get_values_from_str(self, str):
        sig_type = 'val'
        env_signal_names = self.signals_map.keys()
                    
        if str in env_signal_names or str.split('(')[0] in env_signal_names:                        
            sig_val = self.get_sig(str)                        
        elif str.startswith('rho(') or str.startswith('rob('):
            arg_rho = str.split('(')[1][:-1]
            arg_rho = arg_rho.split(',')            
            past_horizon = 0            
            phi = arg_rho[0]
            if len(arg_rho)>1:                          
                past_horizon = -float(arg_rho[1])            
            sig_val = self.get_rob(phi, horizon=past_horizon, online=False)                                                                                                
            sig_type = 'rob'
        elif str.startswith('sat('):
            arg_rho = str.split('(')[1][:-1]
            arg_rho = arg_rho.split(',')
            past_horizon = 0
            phi = arg_rho[0]
            if len(arg_rho)>1:                
                past_horizon = -float(arg_rho[1])
            sig_val = self.get_rob(phi, horizon=past_horizon, online=False)
            sig_val = (sig_val >0).astype(int)         
            sig_type = 'sat'
        else: # try implicit rho(str), i.e., str is a formula name
            sig_val = self.get_rob(str, online=False)                                                                                            
            sig_type = 'rob'                        
        
        return sig_val, sig_type

    def get_rob(self, formula, horizon=0, online=True):
    # Compute robustness signal. If online is true, then 
    # compute it at each time as if future was not known, 
    # otherwise uses all data for all computation
    # MAYBE careful with eval_formula_cfg

        if self.stl_driver.data == []:
            raise ValueError("No data/episode was computed.")
        
        monitor = self.stl_driver.get_monitor(formula)
        
        rob = np.zeros(len(self.stl_driver.data))
        if online:
            monitor.data=[]
        step = 0
        for s in self.stl_driver.data:
            t0 = max(0,step*self.real_time_step-horizon)
            if online:
                monitor.add_sample(s)
            rob[step] = monitor.eval_rob(t0)
            step= step+1
        return rob

    def set_episode_data(self, episode):
        self.reset_monitor()
        self.episode = episode
        observations = self.episode['observations']
        actions = self.episode['actions']
        rewards = self.episode['rewards']            
        rewards_wrapped = self.episode.get('rewards_wrapped',rewards) 
        self.time_step = 0
        self.current_time=0
        
        while self.time_step<len(observations):
            obs = observations[self.time_step]
            action = actions[self.time_step]
            reward_wrapped = rewards_wrapped[self.time_step]                
            s = self.get_sample(obs, action, reward_wrapped)            
            self.stl_driver.add_sample(s)            
            self.time_step +=1
            self.current_time += self.real_time_step

        
    def eval_formula_cfg(self, f_name, f_opt, res=dict()):
    # eval a formula based on f_opt configuration options AT CURRENT STEP

        if f_opt is None:
            f_opt={}

        f_hor = f_opt.get('past_horizon',0)            
        t0 = f_opt.get('t0',max(0, self.current_time-f_hor))
        robs = self.stl_driver.get_online_rob(f_name, t0)
        val = robs[0]
        if res==dict():
            res['estimate_rob'] = np.array(robs[0])
            res['lower_rob'] = np.array(robs[1])
            res['upper_rob'] = np.array(robs[2])                        
            sat = 1 if robs[0]>0 else 0
            res['sat'] = np.array(sat)
        else:
            res['estimate_rob'] = np.append(res['estimate_rob'],robs[0])
            res['lower_rob'] = np.append(res['lower_rob'],robs[1])
            res['upper_rob'] = np.append(res['upper_rob'],robs[2])                        
            sat = 1 if robs[0]>0 else 0
            res['sat'] = np.append(res['sat'],sat)
        
        upper_bound = f_opt.get('upper_bound',np.inf)
        lower_bound = f_opt.get('lower_bound',-np.inf)
        val = max(val, lower_bound)
        val = min(val, upper_bound)

        return val, res
            
        
    def eval_specs_episode(self, episode=None, res=dict(), res_rew_f_list=[], res_eval_f_list=[]):
    # computes different metrics to evaluate an episode. If res contains values already, concatenate
    # returns top level metrics, and robustness and stuff at all steps for reward formulas and eval formulas    
    
        
        if episode is None:
            episode= self.episode            
        else:
            self.episode = episode
                
        rewards = episode['rewards']
        rewards_wrapped = episode.get('rewards_wrapped',rewards)         

        # episode length
        ep_len = len(rewards)        
        res = add_metric(res,'ep_len',ep_len)
        
        # cumulative reward
        ep_rew=0        
        for step in range(0,ep_len):            
            ep_rew +=  rewards[step]
        res= add_metric(res,'ep_rew',ep_rew)

        res_all_ep = dict({'basics':{}, 'reward_formulas':dict(), 'eval_formulas':dict()})            
        res_all_ep['basics']['mean_ep_len'] = np.double(res['ep_len']).mean()
        res_all_ep['basics']['mean_ep_rew'] = res['ep_rew'].mean()
        # maybe a mean mean reward ?

        observations = episode['observations']
        actions = episode['actions']            
        if self.reward_formulas != dict():
            self.reset_monitor()
            self.current_time=0
            self.time_step = 0

            
            res_f = dict()
            for f_name,f_cfg in self.reward_formulas.items():
                res_f[f_name] = []
            
            while self.time_step<len(observations):
                obs = observations[self.time_step]
                action = actions[self.time_step]
                rew_wrapped = rewards_wrapped[self.time_step]                
                s = self.get_sample(obs, action, rew_wrapped)            
                self.stl_driver.add_sample(s)

                for f_name,f_cfg in self.reward_formulas.items():
                    v, _ = self.eval_formula_cfg(f_name,f_cfg)    
                    res_f[f_name] = np.append(res_f[f_name],v)                                        
                    if f_name not in res:
                        res[f_name] = dict() 

                self.time_step +=1
                self.current_time += self.real_time_step
            
            # Synthesize 
            for f_name,f_cfg in self.reward_formulas.items():
                w = f_cfg.get('weight', 1)
                res[f_name] = add_metric(res[f_name], 'mean', w*res_f[f_name].mean())
                
                sum_f = w*res_f[f_name].sum()
                res[f_name] = add_metric(res[f_name], 'sum', sum_f)
                
                num_sat = (res_f[f_name]>0).sum()
                res[f_name] = add_metric(res[f_name], 'num_sat', num_sat)

            res_rew_f_list.append(res_f)            

        # eval formulas - for those, we evaluate off-line, after the trace has been completely computed
            
        if self.eval_formulas != dict():
                        
            self.current_time=0
            self.time_step = 0            

            res_f = dict()
            for f_name,f_cfg in self.eval_formulas.items():
                res_f[f_name] = []
            
            while self.time_step<len(observations):
                
                for f_name,f_cfg in self.eval_formulas.items():
                    v, _ = self.eval_formula_cfg(f_name,f_cfg)    
                    res_f[f_name] = np.append(res_f[f_name],v)                                        
                    if f_name not in res:
                        res[f_name] = dict() 

                self.time_step +=1
                self.current_time += self.real_time_step
            
            # Synthesize: we compute all metrics - maybe we choose in cfg  (TODO)
            for f_name,f_cfg in self.eval_formulas.items():                
                if f_cfg is None:
                    f_cfg = {}
                eval_all_steps = f_cfg.get('eval_all_steps', False)                
                if eval_all_steps:
                    w = f_cfg.get('weight', 1)
                    res[f_name] = add_metric(res[f_name], 'mean', w*res_f[f_name].mean())               
                    sum_f = w*res_f[f_name].sum()
                    res[f_name] = add_metric(res[f_name], 'sum', sum_f)
                    num_sat = (res_f[f_name]>0).sum()
                    res[f_name] = add_metric(res[f_name], 'num_sat', num_sat)                
                else:
                    init_rob = res_f[f_name][0]
                    res[f_name] = add_metric(res[f_name], 'init_rob', init_rob)
                    init_sat = 1 if res_f[f_name][0]>0 else 0
                    res[f_name] = add_metric(res[f_name], 'init_sat', init_sat)                
                
            res_eval_f_list.append(res_f)
                        
            for f_name,f_cfg in self.reward_formulas.items():
              if f_cfg is None:
                    f_cfg = {}                
              if isinstance(res[f_name], dict):                    
                res_all_ep['reward_formulas'][f_name] = dict()
                res_all_ep['reward_formulas'][f_name]['mean_rob'] = res[f_name]['mean'].mean()
                res_all_ep['reward_formulas'][f_name]['mean_num_sat'] = res[f_name]['num_sat'].mean()
                res_all_ep['reward_formulas'][f_name]['mean_sum'] = res[f_name]['sum'].mean()

            num_ep = len(res['ep_len'])
            for f_name,f_cfg in self.eval_formulas.items():
              if f_cfg is None:
                    f_cfg = {}                
              if isinstance(res[f_name], dict):                    
                res_all_ep['eval_formulas'][f_name] = dict()                
                eval_all_steps = f_cfg.get('eval_all_steps', False)
                if eval_all_steps:
                    res_all_ep['eval_formulas'][f_name]['mean_sum'] = res[f_name]['sum'].mean()
                    res_all_ep['eval_formulas'][f_name]['mean_mean'] = res[f_name]['mean'].mean()
                    res_all_ep['eval_formulas'][f_name]['mean_num_sat'] = res[f_name]['num_sat'].mean()
                else:
                    res_all_ep['eval_formulas'][f_name]['ratio_init_sat'] = (res[f_name]['init_sat']>0).sum()/num_ep
                    res_all_ep['eval_formulas'][f_name]['mean_init_rob'] = res[f_name]['init_rob'].mean()

                        

        return res, res_all_ep, res_rew_f_list, res_eval_f_list
    
    


import os
import sys
import argparse

from rlrom.testers import RLTester
from rlrom import utils
from rlrom.trainers import RLTrainer
from pprint import pprint
import numpy as np


def main():
    # rlrom_run [test|train] cfg_main.yml [--cfg-train cfg_train.yml]    
    
    parser = argparse.ArgumentParser(description='Run a configuration file in YAML format for testing or training.')
    parser.add_argument('action', type=str, help='action should be either "test" or "train"')
    parser.add_argument('main_cfg', type=str, default='cfg_main.yml', help='Path to main configuration file in YAML format.')
    parser.add_argument('--cfg_train', type=str, help='Override cfg_train section in main with content of a YAML file.')
    parser.add_argument('--cfg_test', type=str, help='Override cfg_test with content of a YAML file.')
    parser.add_argument('--cfg_specs', type=str, help='Override cfg_specs with content of a YAML file.')
    parser.add_argument('--verbose', type=int, default=0, help='Verbosity level') 
    parser.add_argument('--num-trainings',type=int, default=1, help='Number of repeats of training') #TODO 
    args = parser.parse_args()
    
    # Start with default configuration
    custom_cfg = dict()        
    
    # Load main config file
    try:
        if os.path.exists(args.main_cfg):
            custom_cfg = utils.load_cfg(args.main_cfg)
        else:        
            print(f"Warning: Config file {args.main_cfg} not found.")
    except FileNotFoundError:
        print(f"Error: The main config file '{args.main_cfg}' was not found. Exiting.")
        sys.exit(1)

    # Override with train config if specified
    if args.cfg_train:        
        if os.path.exists(args.cfg_train):
            custom_cfg['cfg_train'] = args.cfg_train
            print(f"Using training config from {args.cfg_train}")
        else:
            print(f"Warning: Training config file {args.cfg_train} not found.")

    # Override with test config if specified
    if args.cfg_test:        
        if os.path.exists(args.cfg_test):
            custom_cfg['cfg_test'] = args.cfg_test
            print(f"Using training config from {args.cfg_test}")
        else:
            print(f"Warning: Training config file {args.cfg_test} not found.")

    # Override with specs config if specified
    if args.cfg_specs:        
        if os.path.exists(args.cfg_specs):
            custom_cfg['cfg_specs'] = args.cfg_specs
            print(f"Using training config from {args.cfg_specs}")
        else:
            print(f"Warning: Training config file {args.cfg_specs} not found.")


    if args.verbose>=1:
        pprint(custom_cfg)
            
    if args.num_trainings:
        num_trainings= args.num_trainings
    else:
        num_trainings = 1
        
    if args.action=='test':
        tester = RLTester(custom_cfg)
        Tres= tester.run_cfg_test()
        tester.print_res_all_ep(Tres)
    elif args.action=='train':    
        trainer = RLTrainer(custom_cfg)
        trainer.train()
    else: 
        print(f'Unrecognized action {args.action}. Should be either "test" or "train"')

if __name__ == "__main__":
    main()


import os
import yaml

supported_envs = []
supported_models = ['ppo', 'a2c', 'sac', 'td3', 'dqn', 'qrdqn', 'ddpg', 'trpo']
cfg_envs = {}

current_dir = os.path.dirname(os.path.abspath(__file__))
dir_cfgs = os.path.join(current_dir, '../rlrom/cfgs/')

list_cfg=  os.listdir(dir_cfgs)
cfg_envs = {}
for c in list_cfg:
    cfg_name,_ = os.path.splitext(c)
    supported_envs.append(cfg_name)
    cfg_full_path = dir_cfgs + c
    with open(cfg_full_path, 'r') as file:
        dict_cfg = yaml.safe_load(file)
    cfg_envs[cfg_name] = dict_cfg


import numpy as np

import gymnasium as gym
from gymnasium.spaces.utils import flatten_space
import stlrom
import rlrom.utils as utils
from rlrom.utils import append_to_field_array as add_metric
from rlrom.wrappers.stl_wrapper import STLWrapper

import rlrom.plots
from bokeh.models.annotations import Title
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show
from bokeh.palettes import Dark2_5 as palette
# itertools handles the cycling
import itertools


def stl_wrap_env(env, cfg_specs):
    driver= stlrom.STLDriver()
    stl_specs_str = cfg_specs.get('specs','')
    if stl_specs_str=='':
        stl_specs_str = 'signal'
        first = True
        for a in cfg_specs.get('action_names',{}):
            if first:
                stl_specs_str += ' '+ a
                first = False
            else:
                stl_specs_str += ','+ a
        for o in cfg_specs.get('obs_names',{}):
            if first:
                stl_specs_str += ' '+ o
                first = False
            else:
                stl_specs_str += ','+ o
        stl_specs_str += ',reward'                                 

    driver.parse_string(stl_specs_str)
    obs_formulas = cfg_specs.get('obs_formulas',{})        
    reward_formulas = cfg_specs.get('reward_formulas',{})
    eval_formulas = cfg_specs.get('eval_formulas',{})
    end_formulas = cfg_specs.get('end_formulas',{})
    BigM = cfg_specs.get('BigM')

    env = STLWrapper(env,driver,
                     signals_map=cfg_specs, 
                     obs_formulas = obs_formulas,
                     reward_formulas = reward_formulas,
                     eval_formulas=eval_formulas,
                     end_formulas=end_formulas,
                     BigM=BigM)
    return env

def make_env_generic(cfg, render_mode='human'):

    env_name = cfg.get('env_name','highway-v0')                   
    env = gym.make(env_name, render_mode=render_mode)
    
    cfg_env = cfg.get('cfg_env',dict())
    if cfg_env != dict():
        env.unwrapped.configure(cfg_env)
      # wrap env with stl_wrapper. We'll have to check if not done already        
    cfg_specs = cfg.get('cfg_specs', None)
            
    if cfg_specs is not None:
        env = stl_wrap_env(env, cfg_specs)
            
    return env

class RLTester:
    def __init__(self,cfg, render_mode='human'):
        
        cfg = utils.load_cfg(cfg)
        self.cfg = cfg        
        self.manual_control = True    
        self.env_name = cfg.get('env_name','highway-v0')                                
        self.env = None
        self.model = None
        self.test_results = []
        self.has_stl_wrapper = cfg.get('cfg_specs', None) is not None
        self.model_use_stl_wrapper = True # if False, model will use observation from the wrapped environment

    def load_model(self):
        cfg_env = self.cfg.get('cfg_env',dict())
        if cfg_env.get('manual_control', False):
            model = 'manual'
            print("INFO: manual_control set to True, stay alert.")            
        else:
            self.manual_control = False
            model_path = self.cfg.get('model_path', './models')
            model_name = self.cfg.get('model_name', 'random')
            if model_path=='huggingface':
                repo_id = model_name
                env_name = self.cfg.get('env_name')
                model = utils.load_model(env_name, repo_id)
            else:
                model_name, _ = utils.get_model_fullpath(self.cfg)
                if model_name=='random':
                    model='random'
                else:
                    print("INFO: Loading model ", model_name)
                    model= utils.load_model(model_name)
        self.model = model

    def _get_action(self, obs):        
        
        if self.manual_control is True or self.model=='random':
            action = self.env.action_space.sample()
        else:
            if self.has_stl_wrapper:  
                if self.model_use_stl_wrapper is False:
                    # agent was trained without stl_wrapper, so we need to use wrapped_obs to predict action
                    obs = self.env.wrapped_obs

            action, _ = self.model.predict(obs)
        return action

    def init_env(self, render_mode=None):
        self.env = make_env_generic(self.cfg, render_mode=render_mode)

    def run_seed(self, seed=None, num_steps=100, render_mode='human', reload_model=False):

        self.init_env(render_mode=render_mode)

        if reload_model:
             self.load_model()
        
        # We actually might want to reload every time to enforce determinism     
        if self.has_stl_wrapper is False:
            episode = {'observations':[], 'actions':[],'rewards':[], 'dones':[]}

        if seed is not None:
            obs, info = self.env.reset(seed=seed)
        else:
            obs, info = self.env.reset()        
        for _ in range(num_steps):    
            action = self._get_action(obs)
            obs, reward, terminated, truncated, info = self.env.step(action)    
            if self.has_stl_wrapper is False:
                episode['observations'].append(obs)               
                episode['actions'].append(action)
                episode['rewards'].append(reward)
                episode['dones'].append(terminated)
            
            if terminated:                
                break    
        
        if self.has_stl_wrapper:
            episode = self.env.episode
        self.env.close()
        return episode

    def run_cfg_test(self):
        cfg_test = self.cfg.get('cfg_test') 
        test_result = dict({'cfg':self.cfg})                        
        if cfg_test is not None:
            init_seed = cfg_test.get('init_seed',0)
            num_ep = cfg_test.get('num_ep',1)
            render = cfg_test.get('render', True)
            num_steps  = cfg_test.get('num_steps', 100)
            reload_model = cfg_test.get('reload_model',False)

            if render:
                render_mode = 'human'
            else:            
                render_mode = None
                cfg_env = self.cfg.get('cfg_env',dict())
                if cfg_env.get('manual_control', False):                    
                    print('WARNING: Manual control was set too True without render, that is dangerous. Setting to False')
                    self.cfg['cfg_env']['manual_control'] = False
                
            test_result = {'episodes':[], 'res':{}}
            num_ep_so_far=0
            if reload_model is False: # we load it here only, otherwise, reload inside run_seed every time
                self.load_model() 

            for seed in range(init_seed, init_seed+num_ep):
                episode = self.run_seed(seed=seed, num_steps=num_steps,render_mode=render_mode, reload_model=False)
                num_ep_so_far+=1
                print('.', end='')
                if num_ep_so_far%10==0:
                    print('|')
                test_result = self.eval_episode(episode, test_result)
            print()
            self.test_results.append(test_result)
            
        return test_result
    
    def eval_episode(self, episode, test_result):
        test_result['episodes'].append(episode)                
        res = test_result.get('res',{})
            
        if self.has_stl_wrapper:
            res_rew_f_list = test_result.get('res_rew_f_list',[])
            res_eval_f_list = test_result.get('res_eval_f_list',[])
            res, res_all_ep, res_rew_f_list, res_eval_f_list  = self.env.eval_specs_episode(res=res,
                                                                                      res_rew_f_list= res_rew_f_list,
                                                                                      res_eval_f_list= res_eval_f_list)
        else: # only episode length and sum reward
            rewards = episode['rewards']
            ep_len = len(rewards)                    
            res = add_metric(res,'ep_len',ep_len)
            ep_rew=0        
            for step in range(0,ep_len):  # computes sum of rewards
                ep_rew +=  rewards[step]
            res= add_metric(res,'ep_rew',ep_rew)            
            
            # synthetic over all past episodes
            res_all_ep = dict({'basics':{}, 'reward_formulas':dict(), 'eval_formulas':dict()})            
            res_all_ep['basics']['mean_ep_len'] = np.double(res['ep_len']).mean()
            res_all_ep['basics']['mean_ep_rew'] = res['ep_rew'].mean()
            test_result['res'] = res
                
        test_result['res_all_ep'] = res_all_ep

        return test_result

    def eval_all_episodes(self, episodes):
        
        test_result = dict({'cfg':self.cfg})                        
        res = dict()        
        res_rew_f_list = []
        res_eval_f_list = []            
        for episode in episodes:                                
                res, res_all_ep, res_rew_f_list, res_eval_f_list  = self.env.eval_episode(episode=episode,
                                                                                          res=res,
                                                                                          res_rew_f_list= res_rew_f_list,
                                                                                          res_eval_f_list= res_eval_f_list)
        test_result['episodes']= episodes
        test_result['res']= res
        test_result['res_all_ep']= res_all_ep
        test_result['res_rew_f_list']= res_rew_f_list
        test_result['res_eval_f_list']= res_eval_f_list
        return  test_result

    def print_res_all_ep(self, test_result):
        res_all_ep = test_result['res_all_ep']
        print('mean_ep_len:', f"{res_all_ep['basics']['mean_ep_len']:.4g}", end=' | ')
        print('mean_ep_rew:', f"{res_all_ep['basics']['mean_ep_rew']:.4g}")
        #for f_name,_ in self.env.reward_formulas.items():
        #    print(f_name+':',f"{res_all_ep['reward_formulas'][f_name]['mean_sum']:.4g}",end=' | ')      
        if self.has_stl_wrapper:
            for f_name,f_cfg in self.env.eval_formulas.items():
                if f_cfg is None:
                    f_cfg = {}                
                is_local_formula = f_cfg.get('eval_all_steps', False)
                if is_local_formula:                                                
                    print(f_name, end=': ')
                    print("sum=",f"{res_all_ep['eval_formulas'][f_name]['mean_sum']:.4g}",end=' | ')
                    print("mean=",f"{res_all_ep['eval_formulas'][f_name]['mean_mean']:.4g}",end=' | ')
                    print("num_sat=",f"{res_all_ep['eval_formulas'][f_name]['mean_num_sat']:.4g}")                
                else:           
                    print(f_name+": ratio_sat=",f"{res_all_ep['eval_formulas'][f_name]['ratio_init_sat']:.4g}")                      
                    
    def find_huggingface_models(self, repo_contains='', algo=''):
        env_name = self.cfg.get('env_name')
        if env_name is None:
            models = []
            models_ids = []
        else:
            models, models_ids = utils.find_huggingface_models(env_name, repo_contains=repo_contains, algo=algo) 
        return models, models_ids

    def get_fig(self, signals_layout, ep_idx=0, test_result=-1):
    # plots stuff in bokeh from episodes in test_results
        
        if isinstance(test_result, int):
            if self.test_results == []: 
                print('PROBLEM: No result computed yet.')
            else:
                test_result = self.test_results[test_result]
        episodes = test_result['episodes']
        current_ep = episodes[ep_idx] 
        self.init_env()
        self.env.set_episode_data(current_ep)
        num_ep = len(episodes)            
        lay = rlrom.plots.get_layout_from_string(signals_layout)
        status = "Plot ok. Hit reset on top right if not visible."            

        figs = []
        colors = itertools.cycle(palette)    

        for signal_list in enumerate(lay):
            f=None
            for signal in signal_list[1]:                
                try: 
                    color=colors.__next__()                                        
                    print(signal.strip())                    
                    if signal.strip().startswith("set_ep_idx(") or signal.strip().startswith("_ep("):            
                        ep_idx = int(signal.split('(')[1][:-1])
                        current_ep = episodes[ep_idx]                         
                        self.env.set_episode_data(current_ep)
                    else: 
                        if f is None:
                            if figs == []:
                                f = figure(height=200)
                            else:
                                f = figure(height=200, x_range=figs[0][0].x_range)
                            figs.append([f])

                        ttime = self.env.get_time()                        
                        labl = signal                                                
                        if num_ep>1:                                                            
                            labl += ', ep '+str(ep_idx)

                        sig_values, sig_type = self.env.get_values_from_str(signal)
                        if sig_type == 'val':
                            f.scatter(ttime, sig_values, legend_label=labl, color=color)
                            f.line(ttime, sig_values, legend_label=labl, color=color)
                        elif sig_type == 'rob':
                            f.step(ttime, sig_values, legend_label=labl, color=color)                        

                        elif sig_type == 'sat':
                            f.step(ttime, sig_values, legend_label=labl, color=color)                        

                except:
                     status = "Warning: error getting values for " + signal
        fig = gridplot(figs, sizing_mode='stretch_width')        

        return fig, status

